! wtellask.h  Ask/Tell conversational library
!             Paolo Lucchesi, 2004
!
! ------------------------------------------------------------------------------
!
! This file is free software, it can be freely used, copied, distribuited and
! modified. Any modified or unmodified copy of this file distribuited as source
! code must acknoledge the original author and must be freely usable, copyable,
! distributable and modifiable.
! There's absolutely no condition or binding on the distribution as compiled
! binary file.
! This software is distribuited as-is, with NO WARRANTY. The author cannot be
! held as responsable for any damage due directly or indirectly to this file.
!
! ------------------------------------------------------------------------------
!
! Library 6/11 or greater needed
!
! How to use:
! Put in your main file the directive
!     include "wtellask";
! two times: after including verblib.h and after including grammar.h
!
! This extension helps the game author to build NPC than can react to commands
! like 'ask npc about topic' or 'tell npc about topic'. Actually this extension
! makes no difference between the two command forms.
!
! Each non-player character should belong to NPC class (which has already the
! animate attribute set), and its life property should not contain the Tell
! and Ask clauses (nor the default clause, as we'll see later). All the other
! actions are handled in the usual way.
! The object belonging to NPC class should have a conversor property pointing
! to an object which is the conversation controller. All the topic options
! available should be children of this conversor, so you can add and subract
! dialog options (even if you won't use this feature very often).
! The conversor object contains, in its description properties, the dialog
! fragment used to answer to the 'talk with npc' (with no topics) command.
! This can be a string or a routine (which can react differently, depending
! on the game state).
! The description properties can also be a list of strings or routines that
! will be used sequentially for each 'talk to npc' command; the last element
! will be repeated for every further command. The NPC's cnumber property keeps
! count of how many times the 'talk to npc' command has been used.
! Furthermore the conversor object can contain, in the cant_go property, the
! dialog fragment used when no correspondence is found for the requested
! topic (when the npc doesn't know anything about the topic, for example 
! "Sorry, I don't know anything about that...").
! Finally if the conversor object has a before property, this property is
! run before doing anything else (obviously just in response to a Tell or Ask
! action); if the before returns true, there will be no further response to
! the command.
!
! Topics should belong to Topic Class. 
! The name list of each topics contains the names with which the player can
! refer to the topic.
! The topic's description properties contains the dialog fragment used, and
! it can be a string or a routine; in the latter case the routine can
! obviously do anything and react differently under different conditions.
! The description property can also be a list of strings or routines that
! will be used sequentially every time that the topic is requested. The last
! element will be repeated for every further command (but see the metatopic
! property). The cnumber property keeps track of how many times the topic
! has been requested.
! Optionally, each topic object can have an after properties, which is run
! just after showing the dialog fragment.
! Also, each topic object can have a before properties, which is run just
! before showing the dialog fragment. If the before properties returns true,
! then the dialog is not shown (the description and after properties are
! not evaluated).
! Any object belonging to the Topic class can have the locked attribute set.
! If this happens, this extension assume that the player doesn't know anything
! about that topic, and don't allow it. But the before properties is run
! even if the locked attribute is set (and can modify the locked attribute).
!
! PAY ATTENTION: in the NPC life properties you cannot use the default: case
! anymore. You should use the lifedefault properties instead.
!
! Every object belonging to the Topic class can have a metatopic properties
! pointing to another object of Topic class; this latter object should 
! contain general elements regarding a certain topic, indipendently from the
! NPC the player is talking with. So, if a metatopic is defined, there are
! three effects:
! - The list of the names that refers to the topic is taken from the name
!   properties both of main Topic object and of its metatopic
! - Even if the metatopic object has the locked attribute set, the dialog
!   topic is not available
! - The list items of the metatopic's description property (if it exists) are
!   added to the main topic description list (so, if the main topic hasn't
!   a description property, the metatopic description is used instead).
!
! This library adds the 'talk about topic' (also 't topic') and 'ask about
! topic' commands, which automatically refer to the NPC the player is already
! talking to (if he/she exist).
! Usually the actual talker is the last NPC that the player has talked/asked
! to (if he/she's been always in scope), and his/her id is contained in the
! talkingwith variable.
! The game programmer can make a conversation with a certain NPC start by
! calling npc.talkto(flag) or talker(npc, flag), where png is the character
! who should became the new actual talker, and flag is a true/false variable
! stating if the conversation should be started anyway even if the player is
! already talking with someone else.
! Furthermore, the silenceturns variable contains the number of turns
! since the last conversation command.
!
! Example
!
! Topic metapalace
!   with name 'palace' 'library' 'building';
!   
! Topic metabook
!   with name 'book' 'black',
!   has locked;
!   
! NPC -> librarian "librarian"
!   with name 'girl' 'young' 'woman' 'librarian',
!     description "Still young and pretty.",
!     conversor girltalk, ! <--- conversation controller
!     Life [;
!       Answer: "You don't believe she can help you.";
!       Show, Give: 
!             print_ret "The girl looks at ", (the)noun, ", but is unimpressed.";      
!       ThrowAt, Attack: "Did she something wrong to you?";
!     ],
!     lifedefault [; "The girl looks you with curiosity"; ], ! <--- al posto di Default:
!   has female;
!
! Object girltalk 
!   with description "~Hi, how are you?~^
!                    ~I'm fine... and you?~",   ! <--- answer to 'talk with girl'
!        cant_go     "~Sorry, I don't know anything about that.~";   ! <--- no topic found
!
! Topic -> 
!    with metatopic metapalace,
!    description [;
!      give metabook ~unlocked; ! <- Now we can talk about the book
!      "~What can you tell me about the library?~^
!       ~Built in the 1666. In it's collection are really rare volumes,
!        like the black book on that shelf.~";
!     ];
!             
! Topic -> 
!    with name 'volume',
!    metatopic metabook,
!    description [;
!        if (self hasnt general) { ! <--- Changing dialog
!            give self general;
!            "~Can you tell me something about that book with the black cover?~^
!              It's the book of Eibon. It's really dangerous.~";
!        }
!        "~About that black book...~^
!         ~As I've already told you, leave it alone.~";
!    ];
!
! Topic ->
!    with name 'picture' 'painting' 'portrait',
!    ! Phrases list example
!    description "~Who's the man in that portrait?~^
!                 ~Count Parnias. He was the first library owner.~"
!                "~Do you know who has made this painting?~^
!                 ~He was called Thomas Malwing. I don't know anything about him.~"
!                "You don't have anything more to ask about the portrait.";
!
! ------------------------------------------------------------------------------

! ----
! First inclusion

#Ifndef LIBRARY_GRAMMAR;

Global talkingwith=NULL;
Global silenceturns=0;
Global wtopicword=0;

Property cnumber;
Property act;
Property talkto;
Property metatopic;

Class NPC
  with 
    number 0, cnumber 0,
	life [wd i x mtx j;
      Tell,Ask:
        if (~~(self provides conversor)) rfalse;
        talkingwith = self; silenceturns=0;
        if ((self.conversor) provides before) {
            x = (consult_words==0 && action==##Tell);
            if ((self.conversor).before(x)) rtrue;
        }
        if (consult_words==0 && action==##Tell) {
            if ((self.conversor) provides description) {
			  j = (self.conversor).#description/WORDSIZE;
			  if (j <= 0) rfalse;
			  self.cnumber++; 
			  if (self.cnumber < 1) self.cnumber=1;
			  if (self.cnumber > j) self.cnumber = j;
              x=(self.conversor).&description-->(self.cnumber-1);
              if (metaclass(x) == Routine) x(); else print (string) x, "^";
              rtrue;
            }
            rfalse;
        }
        wn=consult_from;
        for (i=0: i<consult_words: i++) {
            wd=NextWord(); wtopicword=wd;
            objectloop (x in self.conversor) {
                if (x ofclass Topic) {
                    mtx = NULL;
                    if (x provides metatopic) mtx = x.metatopic;
                    if (~~(mtx ofclass Topic)) mtx=NULL;
                    if (mtx ~= NULL or 0) {
                        for (j=0: j<((mtx.#name) / 2): j++) {
                            if ( ( wd ) == ( (mtx.&name-->j) ) ) {
                                dowdialog(x, mtx);
                                rtrue;
                            }
                        }
                    }
                    for (j=0: j<((x.#name) / 2): j++) {
                        !print i, " ", j, " : ", wd, " ", x&name-->j;
                        if ( ( wd ) == ( (x.&name-->j) ) ) {
                            dowdialog(x, mtx);
                            rtrue;
                        }
                    }
                }
            }
        }
        if ((self.conversor) provides cant_go) {
            PrintOrRun((self.conversor), cant_go, 0);
            rtrue;
        }
	  Greet: 
          ! <<Tell self>>;
          if (~~(self provides conversor)) rfalse;
          talkingwith = self; silenceturns=0;   
          if ((self.conversor) provides before)               
              if ((self.conversor).before(true)) rtrue;
          if ((self.conversor) provides description) {
			  j = (self.conversor).#description/WORDSIZE;
			  if (j <= 0) rfalse;
			  if (self.cnumber < 1) self.cnumber=1;
			  if (self.cnumber > j) self.cnumber = j;
              x=(self.conversor).&description-->0;
              if (metaclass(x) == Routine) x(); else print (string) x, "^";
			  rtrue;          
          }
          rfalse;
      default:
          if (self provides lifedefault) {
              PrintOrRun(self, lifedefault, 0);
              rtrue;
          }
  ],
  talkto [force;
      if ((talkingwith == NULL || force) && (self provides conversor)) 
          { talkingwith = self; silenceturns=0; }
  ],
  has animate;
  
Class Topic
  with number 0,
       cnumber 0,
       act [x;
           if (self provides metatopic) x = self.metatopic; else x = 0;
           dowdialog(self, x);
       ];       

Object wtellaskfakeobject
    with metatopic, lifedefault,
      __warningsquasher [;  talker(self,false); ];
      
[ dowdialog tp mtp rval nt nmt x;
    if (tp provides before) if (tp.before()) rtrue;
    if (tp has locked) "You don't know anything about that.";
    if (mtp ~= NULL or 0) if (mtp has locked) "You don't know anything about that.";
    rval=false;
    nt = 0; nmt = 0;
	if (tp provides description) nt = tp.#description/WORDSIZE;
	if (mtp ~= NULL or 0) if (mtp provides description) nmt = mtp.#description/WORDSIZE;
    if ((nt+nmt) > 0) {
	  tp.cnumber++;
	  ! print nt, " + ", nmt, " ( ", tp.cnumber, " )^";
	  if (tp.cnumber < 1) tp.cnumber=1;
	  if (tp.cnumber > (nt+nmt)) tp.cnumber = (nt+nmt);
	  if (tp.cnumber > nt) x=mtp.&description-->(tp.cnumber-nt-1);
	   else                x=tp.&description-->(tp.cnumber-1);
      if (metaclass(x) == Routine) rval=x(); else { print (string) x, "^"; rval = true; }
    }
    if (tp provides after)
        rval=tp.after();
    rtrue;
];

Object "(wtellask)" LibraryExtensions
  with ext_initialise [;
    TalkInit();
    rtrue;
  ];

[ TalkInit; ! Left for legacy;
    talkingwith = NULL; 
    StartDaemon(talkdaemon);
];

Object talkdaemon
  with daemon [;
    if (talkingwith ~= NULL)
        if (~~TestScope(talkingwith)) talkingwith=NULL;   
    if (talkingwith) silenceturns++;     
];

[ talker id force;
      if ((talkingwith == NULL || force || id == NULL) && (id provides conversor)) 
          { talkingwith = id; silenceturns=0; }
];

! Routine per topic
[KeepTalkSub;
  if (talkingwith==NULL) "You're not talking with anyone, right now.";
  <<Tell talkingwith noun>>;
];

[KeepAskSub;
  if (talkingwith==NULL) "You're not talking with anyone, right now.";
  <<Ask talkingwith noun>>;
];

! Greet
[GreetSub;
  if (RunLife(noun,##Greet)~=0) rfalse;
  L__M(##Greet,1,noun);
]; 

#Endif;


! ----
! Second inclusion

#Ifdef LIBRARY_GRAMMAR;

Extend 'tell'
                * 'about' topic                  -> KeepTalk;

Extend 'ask'
                * 'about' topic                  -> KeepAsk;

Verb 'topic' 't//'
                * topic                          -> KeepTalk;
				
Verb 'greet' 'hail' 'salute'
				*                                -> WaveHands
                * creature                       -> Greet;           

#Endif;

